#! /usr/bin/python3
#

import os
import sys
import sysconfig
import subprocess

verbosity = int(os.getenv('RPM_SCRIPTS_DEBUG', 0))
prog = 'python3.prov'
extension_suffix = sysconfig.get_config_var('EXT_SUFFIX')
if extension_suffix is None :
	extension_suffix = sysconfig.get_config_var('SO')

def normpath(f) :
    if not f.startswith('/') :
        f = os.getcwd() + f
    return os.path.normpath(f)

def get_buildroot() :
    buildroot = os.getenv('RPM_BUILD_ROOT','')
    if buildroot in ['','/'] :
        return ''
    return normpath(buildroot)

buildroot = get_buildroot()
br_len = len(buildroot)
def strip_buildroot(p) :
        if p.startswith(buildroot+'/') :
                return p[br_len:]
        return p

def get_path() :
    "get the list of standard python directories"
    # we assume that the first sys.path element is "" which stands for current directory,
    # and that sys.path entries are already normalized.
    path = sys.path[1:]
    if br_len > 0 :
        # it is possible that sys.path already has RPM_BUILD_ROOT entries
        # e.g. if we build python itself and RPM_PYTHON is redefined
        path = [ strip_buildroot(p) for p in path ]
    # now we can prepend RPM_BUILD_ROOT
    if br_len > 0 :
        br_path = [ buildroot + p for p in path ]
        path = br_path + path
    return path

def get_pth_path(sitedirs) :
    "get the list of additional python directories from site-packages/*.pth"
    # cf. site.py:addsitedir()
    def glob_pth(dir) :
        try :
            files = os.listdir(dir)
        except :
            return []
        return [ f for f in files if f.endswith('.pth') ]
    def read_pth(f) :
        try :
            fp = open(f)
        except :
            return []
        subdirs = []
        for line in fp :
            if line.startswith("#"):
                continue
            if line.startswith("import"):
                continue
            line = line.rstrip()
            if len(line) == 0:
                continue
            subdirs.append(line)
        return subdirs
    # okay, here we go
    pth_path = []
    for sitedir in sitedirs :
        for pth in glob_pth(sitedir) :
            for subdir in read_pth(sitedir+'/'+pth) :
                dir = normpath(sitedir+'/'+subdir)
                pth_path.append(dir)
        pth_path = [ strip_buildroot(p) for p in pth_path ]
        br_pth_path = [ buildroot + p for p in pth_path ]
        pth_path = br_pth_path + pth_path
    #print >>sys.stderr, "%s: pth_path=" % prog, pth_path
    return pth_path

path = get_path()
path += get_pth_path([ p for p in path if p.endswith('/site-packages') ])
# it is essential to sort path entries descending by their length,
# so that "most specific" entries are tried first; e.g.
# /usr/lib/python2.5/site-packages should go before /usr/lib/python2.5
def psort(s):
    return len(s)
path.sort(key=psort)
path.reverse()
# handle RPM_PYTHON3_LIB_PATH which can be exported with rpm macros
rpm_path = os.getenv('RPM_PYTHON3_PATH','').split() + os.getenv('RPM_PYTHON3_COMPILE_INCLUDE','').split()
# Processing .pth (as with `path`) seems not to be important for these
# non-standard paths, because ALT Sisyphus' contents_index doesn't
# contain any *.pth files outside the standard paths.
rpm_path = [p for p in rpm_path if p.startswith('/')]
# If a standard Provides can be generated, do not generate the non-standard one:
rpm_path = [ p for p in rpm_path if not (p in path) ]
rpm_path.sort(key=psort)
rpm_path.reverse()

path = [ p for p in path if os.path.exists(p) ]
# We do not prepend `buildroot' to `rpm_path' (as with `path'),
# because it must be prepended always (and sometimes stripped --
# to generate the custom "path-set" version for the corresponding Provides).
#
# Explanation: Since we decided not to process *.pth inside
# `rpm_path`, we do not need to read anything from the corresponding
# paths in the build system in order to generate Provides.
rpm_path = [ p for p in rpm_path if os.path.exists(buildroot + p) ]

# debug
if __name__ == '__main__' and verbosity > 1 :
    print("%s: path=" % prog, path, file=sys.stderr)
    print("%s: rpm_path=" % prog, rpm_path, file=sys.stderr)

def explode(f) :
    "split filename into std-dir + subdir/basename"
    longest_p_containing_f = None
    for p in path :
        if f.startswith(p+'/') :
            longest_p_containing_f = p
            break
    for p in rpm_path :
        if f.startswith(buildroot+p+'/') :
            l = br_len+len(p)
            if longest_p_containing_f and (len(longest_p_containing_f) >= l) :
                return None,longest_p_containing_f,f[len(longest_p_containing_f)+1:]
            else :
                return p,f[:l],f[l+1:]
    if longest_p_containing_f :
        return None,longest_p_containing_f,f[len(longest_p_containing_f)+1:]
    else :
        return None, None, None

def prov1db(d,b) :
    "list of modules provided by the file, split into std-dir + subdir/basename"
    b = b.replace(extension_suffix,".so")
    # Python syntax does not allow modules with dashes.
    # Also note that this will exclude e.g. b="site-packages/foo.py".
    if b.find("-") >= 0 :
        return
    m,e = os.path.splitext(b)
    if not e in ['.py','.so'] :
        return
    if not '/' in m :
        # the file is a simple module
        if m == '__init__' :
            return
        prov = [ m ]
        if e == '.so' and len(m) > 6 and m[-6:] == 'module' :
            # e.g. rpmmodule.so provides rpm
            prov.append(m[0:-6])
        return prov
    # hierarchical module
    parts = m.split('/')
    if parts[-1] == '__init__' :
        # e.g. foo/bar/__init__.py provides foo.bar
        return [ '.'.join(parts[:-1]) ]
    # for/bar.py provides foo.bar if foo/__init__.py is present
    if os.path.exists(d+'/'+os.path.dirname(b)+'/__init__.py') :
        m = '.'.join(parts)
        prov = [ m ]
        if e == '.so' and len(m) > 6 and m[-6:] == 'module' :
            prov.append(m[0:-6])
        return prov

def prov1(f) :
    "returns list of modules provided by the file"
    nonstd,d,b = explode(f)
    if not d :
        return
    prov = prov1db(d,b)
    if prov :
            prov = [ (nonstd,x) for x in prov ]
    # Sometimes a module can be loaded by its full name, e.g. PIL.Image,
    # while PIL.pth adds PIL subdir to std-dirs, so that the module can be
    # loaded simply as Image.  We provide limited support for this feature.
    # Note that, since PIL.pth provides the longest std-dir, the b=Image case
    # has just been handled.  Now we check d to see if we need to recombine
    # std-dir + subdir/basename and run prov1db(d,b) again.
    while os.path.exists(d+'/__init__.py') :
        b = os.path.basename(d)+'/'+b
        d = os.path.dirname(d)
        xprov = prov1db(d,b)
        if xprov :
            if d in path :
                xprov = [ (None,x) for x in xprov ]
            elif strip_buildroot(d) in rpm_path :
                xprov = [ (strip_buildroot(d),x) for x in xprov ]
            else :
                xprov = []
            if prov :
                prov += xprov
            else :
                prov = xprov
    return prov

def verbose(msg) :
    if verbosity > 0 :
        print("%s: %s" % (prog,msg), file=sys.stderr)

def rpm_format_strset(s) :

    # Borrowed from lib.req:
    # For the number of symbols in the range 2^{m-1}..2^m-1,
    # use m+10 bits per symbol.  Upper bound on the error rate
    # is 2^{-10} (about 0.1%).
    import math
    bpp = math.log(len(s)) / math.log(2) + 11

    mksetP = subprocess.Popen(['/usr/lib/rpm/mkset', '%d' % bpp],
                              stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE,
                              universal_newlines=True)
    return mksetP.communicate(input='\n'.join(s))[0].strip()

def provides(f,m,nonstd) :
    prefix = "python%u" % sys.version_info[0]
    prov = "%s(%s)" % (prefix,m)
    if not nonstd :
        verbose("%s provides %s" % (f,prov))
        print(prov)
    else :
        print("%s provides %s (under non-std path %s)" % (f,prov,nonstd), file=sys.stderr)
        print("%s = %s" % (prov,rpm_format_strset([nonstd])))
    # For the 1st stage of the "cosmetic" renaming of reqs/provs in
    # Sisyphus, provide the old-style python3.3(*), so that the
    # requirements of old modules are satisfied when they are used
    # during the transition.
    #
    # NB: Remove this hack on the 3rd stage of the transition
    # (when the autoreqs will have been renamed).
    #
    # (On the 1st stage, the renaming of the autoreqs is not done--in
    # order to simplify the usage of the newly built modules during the
    # transition.)
    #
    # The code below is not generic: it is tied to the "3.3" version,
    # because we have a pile of old-style provs/reqs of this specific
    # form in Sisyphus now. Introducing "3.5" old-style provs would
    # make no sense. Anyway, this is just a cosmetic renaming; the
    # actual ability of a module to be "run" by a specific version of
    # python is tracked by other deps: either the dependency on
    # python3.3/site-packages/ in the old situation, or the planned
    # dependency on libpython.
    oldprefix = "python3.3"
    if (prefix != oldprefix) and not nonstd :
        prov = "%s(%s)" % (oldprefix,m)
        verbose("%s provides (for the transition from old-style python3.3) %s" % (f,prov))
        print(prov)

def main(files, provides=provides, verbose=verbose) :
    for f in files :
        prov = prov1(f)
        if prov :
            for nonstd,p in prov :
                provides(f,p,nonstd)
                # TODO: do not print provs at nonstd paths if there is a module
                # with the same name at a std path (prefer the externally visible
                # namespace). Example: could help with poezio. But such situation
                # is not nice in general and should be avoided in Python source
                # code; and the priority of implementing this feature here is low.
        else :
            verbose("%s provides nothing" % f)

if __name__ == '__main__':
    files = sys.argv[1:] or [ x.strip() for x in sys.stdin.readlines() ]
    main(files)

# ex: set ts=8 sts=4 sw=4 et:
